;-----------------------------------------------------------------------;
; Low-Lewel Routines for Xpl0re v1.0, (x) 1999 by PREDAt0R		;
;-----------------------------------------------------------------------;

;
; GetKernel32Base routine retrieves kernel32.dll image base
;
; Entry:
;	 lpK32Entry - is an entry point somewhere at the kernel32.dll code,
;	 	      for exaple it can be pointer to terminating process
;		      code, which winloader puts at the top of stack before
;		      will loaded program will start
;
; Exit:
;	 eax	    - contains kernel32 image base 
;		      or null, if K32Entry wasn't entry at the kernel32
;

GetKernel32Base	proc	lpK32Entry:dword
		uses	ebx,esi

; search for kernel32.dll export table rva/size

	mov	esi,dword ptr [lpK32Entry]
	mov	eax,'EP'		; PE-header signature: 00004550h

search_for_pe_sign:
	dec	esi
	cmp	dword ptr ds:[esi],'EP'	; look for PE-header signature bytes
	jne	search_for_pe_sign
	cmp	word ptr ds:[esi+5ch],2	; check one more value to be ensure
	jne	search_for_pe_sign	; this is really PE-header (NT subsys)

; get kernel32.dll image base

	mov	eax,dword ptr [esi+34h]	; get kernel32.dll image base

; check if module is really kernel32

	mov	ebx,dword ptr [esi+78h]	; calculate address of export table
	add	ebx,eax
	mov	ebx,dword ptr [ebx+0ch]	; calculate address of dll name str
	add	ebx,eax
	cmp	dword ptr [ebx],'NREK'	; KERNEL32 ?
	jne	bad_kernel32
	cmp	dword ptr [ebx+4],'23LE'
	je	kernel32_found

bad_kernel32:
	sub	eax,eax			   ; clear eax if error

kernel32_found:
	ret

GetKernel32Base	endp

;
; GetFuncAddress routine returns the address of the specified exported
; by dynamic-link library (DLL) function
;
; Entry:
;	 lpModuleBase - pointer to a memory-mapped DLL image base
;	 lpFuncName   - pointer to zero-terminated ASCIIZ string
;		        of DLL's exported function name
;
; Exit:
;	 eax	      - linear address of specified function or null
;		        if it cannot be retrieved
;

GetFuncAddress	proc	lpModuleBase,lpFuncName:dword
		uses	ebx,ecx,esi,edi

	mov	esi,dword ptr [lpModuleBase]
	add	esi,dword ptr [esi+3ch]	; address of NE (PE) header
	mov	ebx,dword ptr [esi+34h]	; ebx = image base of DLL module
	mov	esi,dword ptr [esi+78h] ; esi = Export Dir Table RVA
	or	esi,esi			; check if EDT is present
	jz	no_export_table

	add	esi,ebx			; esi = Export Dir Table address
	mov	eax,dword ptr [esi+24h] ; eax = Ordinal Table RVA
	or	eax,eax
	jz	no_export_table

	push	dword ptr [esi+1ch]	; (+1) save Address Table RVA
	push	dword ptr [esi+10h]	; (+2) save Ordinal Base value
	add	eax,ebx			;      eax = Ordinal Table address
	push	eax			; (+3) save it
	mov	ecx,dword ptr [esi+18h]	;      ecx = number of Name Pointers
	mov	esi,dword ptr [esi+20h] ;      esi = Name Table Pointers RVA
	add	esi,ebx			;      esi = NTPaddress
	push	esi			; (+4) save it

	cld
search_specified_export:
	lodsd				; get next pointer (RVA format)
	add	eax,ebx			; convert into address format
	xchg	edi,eax			; edi = pointer to ASCIIZ-string name
	push	esi			; (+5) save esi
	mov	esi,dword ptr [lpFuncName]

compare_strings:			; comapre two strings
	lodsb
	scasb
	jnz	strings_are_not_equal
	or	al,al
	mov	eax,dword ptr [edi]
	jnz	compare_strings

strings_are_equal:			; if we have found equal string:
	pop	eax			; (-5) eax = addr. of pointer 2 name
	pop	edi			; (-4) edi = address of NTP
	sub	eax,edi
	shr	eax,1			; eax = index to access Ordinal Table
	pop	edi			; (-3) edi = address of Ordinal Table
	movzx	eax,word ptr [edi+eax]	;      eax = ordinal value of function
	pop	edi			; (-2) edi = Ordinal Base values
	sub	eax,edi			;      eax = index to access AT
	pop	edi			; (-1) edi = RVA of Address Table
	add	edi,ebx			;      edi = address of Address Table

	mov	eax,dword ptr [edi+4*eax] ; eax = address of exported function
	add	eax,ebx
	ret

strings_are_not_equal:			; repeat with next RVA 'till we
	pop	esi			; will find equal string (-5)
	loopd	search_specified_export
	add	esp,4*04		; (-4), (-3), (-2), (-1)

no_export_table:			; no EDT or specified function was
 	sub	eax,eax			; found
	ret

GetFuncAddress	endp

;
; MapDll routine maps specified DLL at the current context
;
; Entry:
;	 lpK32Entry   - is an entry point somewhere at the kernel32.dll code,
;	 	        for exaple it can be pointer to terminating process
;		        code, which winloader puts at the top of stack before
;		        will loaded program will start
;	 lpModuleName - pointer to a DLL ASCIIZ string module name
;
; Exit:
;	 eax	      - DLL handle or null if DLL cannot be mapped
;


MapDll		proc	lpK32Entry,lpModuleName:dword


	call	GetKernel32Base,[lpK32Entry]; get Kernel32 base at eax
	or	eax,eax
	jz	cannot_map_dll

	call	GetFuncAddress,eax,ofs _LLA ; get linear entry point into
	or	eax,eax			    ; KERNEL32!LoadLibraryA func
	jz	cannot_map_dll

	call	eax,[lpModuleName]

cannot_map_dll:
	ret

_LLA	db	"LoadLibraryA",0

MapDll		endp

;
; UnmapDll routine unmaps specified DLL from the current context
;
; Entry:
;	 lpK32Entry - is an entry point somewhere at the kernel32.dll code,
;	 	      for exaple it can be pointer to terminating process
;		      code, which winloader puts at the top of stack before
;		      will loaded program will start
;	 hLibModule - DLL module handle retrieved by MapDll (LoadLibraryA)
;
; Exit:
;	 eax	    - null if DLL cannot be unmapped or nonzero in other case
;


UnmapDll	proc	lpK32Entry,hLibModule:dword


	call	GetKernel32Base,[lpK32Entry]; get Kernel32 base at eax
	or	eax,eax
	jz	cannot_unmap_dll

	call	GetFuncAddress,eax,ofs _FL  ; get linear entry point into
	or	eax,eax			    ; KERNEL32!LoadLibraryA func
	jz	cannot_unmap_dll

	call	eax,[hLibModule]

cannot_unmap_dll:
	ret

_FL	db	"FreeLibrary",0

UnmapDll	endp

